from turf.distance import distance

from turf.utils.error_codes import error_code_messages
from turf.utils.exceptions import InvalidInput


def square(bbox):
    """
    Takes a bounding box and calculates the minimum square bounding box that
    would contain the input.

    :param bbox: bounding box extent in [minX, minY, maxX, maxY] order

    :return: a square surrounding bbox

    """
    if not isinstance(bbox, list) or len(bbox) != 4:
        raise InvalidInput(error_code_messages["InvalidBoundingBox"])

    west = float(bbox[0])
    south = float(bbox[1])
    east = float(bbox[2])
    north = float(bbox[3])

    horizontal_distance = distance([west, south], [east, south])
    vertical_distance = distance([west, south], [west, north])

    if horizontal_distance >= vertical_distance:
        vertical_midpoint = (south + north) / 2
        bounding_box = [
            west,
            vertical_midpoint - ((east - west) / 2),
            east,
            vertical_midpoint + ((east - west) / 2),
        ]
    else:

        horizontal_midpoint = (west + east) / 2
        bounding_box = [
            horizontal_midpoint - ((north - south) / 2),
            south,
            horizontal_midpoint + ((north - south) / 2),
            north,
        ]

    return bounding_box



from typing import Dict, List

import numpy as np

from turf.helpers import convert_length, degrees_to_radians
from turf.helpers import earth_radius
from turf.invariant import get_coords_from_features


def rhumb_distance(origin, destination, options: Dict = None) -> float:
    """
    Calculates the rhumb distance between two Points. Units are defined in helpers._units

    # https://en.wikipedia.org/wiki/Rhumb_line

    :param start: starting point [lng, lat] or Point feature
    :param end: ending point [lng, lat] or Point feature
    :param options: dictionary with units as an attribute.
                    Units are defined in helpers._units

    :return: distance between the 2 points

    """
    if not isinstance(options, dict):
        options = {}

    origin = get_coords_from_features(origin, ["Point"])
    destination = get_coords_from_features(destination, ["Point"])

    # compensate the crossing of the 180th meridian (https://macwright.org/2016/09/26/the-180th-meridian.html)
    # solution from https://github.com/mapbox/mapbox-gl-js/issues/3250#issuecomment-294887678
    if (destination[0] - origin[0]) > 180:
        destination[0] -= 360
    elif (origin[0] - destination[0]) > 180:
        destination[0] += 360

    distance_in_meters = calculate_rhumb_distance(origin, destination)

    distance = convert_length(
        distance_in_meters, "meters", options.get("units", "kilometers")
    )

    return distance


def calculate_rhumb_distance(
    origin: List, destination: List, radius: float = None
) -> float:
    """
    Calculates the rhumb distance between two Points.

    # https://en.wikipedia.org/wiki/Rhumb_line

    :param origin: starting point [lng, lat]
    :param destination: ending point [lng, lat]
    :param radius: radius of the earth

    :return: distance between the 2 points in meters
    """
    if not radius:
        radius = earth_radius

    phi_1 = degrees_to_radians(origin[1])
    phi_2 = degrees_to_radians(destination[1])
    delta_phi = phi_2 - phi_1

    delta_lambda = degrees_to_radians(abs(destination[0] - origin[0]))

    # if dLon over 180° take shorter rhumb line across the anti-meridian:
    if delta_lambda > np.pi:
        delta_lambda -= 2 * np.pi

    # on Mercator projection, longitude distances shrink by latitude; q is the 'stretch factor'
    # q becomes ill-conditioned along E-W line (0/0); use empirical tolerance to avoid it

    delta_psi = np.log(np.tan(phi_2 / 2 + np.pi / 4) / np.tan(phi_1 / 2 + np.pi / 4))

    if abs(delta_psi) > 10e-12:
        q_1 = delta_phi / delta_psi
    else:
        q_1 = np.cos(phi_1)

    # distance is pythagoras on 'stretched' Mercator projection
    delta = np.sqrt(delta_phi * delta_phi + q_1 * q_1 * delta_lambda * delta_lambda)
    # angular distance in radians
    distance = delta * radius

    return distance



import math
from typing import List, Dict, Union
import json

from turf.distance._distance import calculate_radians_distance
from turf.helpers import degrees_to_radians, radians_to_degrees
from turf.helpers import point, line_string
from turf.helpers import units_factors
from turf.helpers import Feature, LineString, Point
from turf.utils.exceptions import InvalidInput
from turf.utils.error_codes import error_code_messages


class Coordinate:
    def __init__(self, lon: float, lat: float, decimals=6):

        self.lon = lon
        self.lat = lat
        self.x = degrees_to_radians(lon)
        self.y = degrees_to_radians(lat)
        self.decimals = decimals

    @property
    def coords(self) -> List[Union[float, float]]:
        return self._rnd([self.lon, self.lat])

    @property
    def point(self) -> Point:
        return point(self._rnd([self.lon, self.lat]))

    def to_dict(self):
        d = {"lat": self.lat, "lon": self.lon}
        return d

    def __repr__(self):
        return f"Position({self.lon},{self.lat})"

    def __str__(self):
        return json.dumps(self.to_dict())

    def _rnd(self, x):
        return [round(i, self.decimals) for i in x]


class GreatCircle:
    def __init__(self, start, end, properties):

        self.start = Coordinate(start[0], start[1])
        self.end = Coordinate(end[0], end[1])

        self._check_antipodal()

        self.properties = properties
        self.distance = self._get_distance()
        self.arc_coordinates = self._calculate_arc_coordinates()

    def _check_antipodal(self):
        """
        Antipodes are coordinates that are diametrically opposite to each
        other on Earth. Antipodal points are connected by an infinite number of
        great circles
        """
        dif_lon = self.start.lon - self.end.lon
        dif_lat = self.start.lat - self.end.lat

        if (dif_lat == 0) and ((abs(dif_lon % 360) - 180) == 0):
            raise InvalidInput(
                error_code_messages["InvalidGreatCirclePoints"](
                    self.start.coords, self.end.coords
                )
            )

    @property
    def linestring(self) -> Union[Feature, Dict]:
        return line_string(self.arc_coordinates, self.properties)

    def _get_distance(self) -> float:
        """
        Gets the distance between start and end

        :return: distance_radians
        """
        dif_lon = self.start.x - self.end.x
        dif_lat = self.start.y - self.end.y

        distance = calculate_radians_distance(
            dif_lon, dif_lat, self.start.y, self.end.y
        )

        return distance

    def _get_intermediate_coord(self, fraction: float) -> List[Union[float, float]]:
        """
        Calculates the intermediate point on a great circle line
        http://www.edwilliams.org/avform.htm#Intermediate

        :param fraction: input fraction of the whole great circle
        :return: a tuple of cordinates
        """
        A = math.sin((1 - fraction) * self.distance) / math.sin(self.distance)
        B = math.sin(fraction * self.distance) / math.sin(self.distance)

        x = A * math.cos(self.start.y) * math.cos(self.start.x) + B * math.cos(
            self.end.y
        ) * math.cos(self.end.x)

        y = A * math.cos(self.start.y) * math.sin(self.start.x) + B * math.cos(
            self.end.y
        ) * math.sin(self.end.x)

        z = A * math.sin(self.start.y) + B * math.sin(self.end.y)

        lat = radians_to_degrees(
            math.atan2(z, math.sqrt(math.pow(x, 2) + math.pow(y, 2)))
        )
        lon = radians_to_degrees(math.atan2(y, x))

        return Coordinate(lon, lat).coords

    def _calculate_arc_coordinates(self) -> LineString:
        """
        Calculates intermediate points on a great circle line

        :param n_points: amount of intermediate points on the great circle
        :return: list of coordinates
        """
        coordinates = []
        n_points = self.properties.get("npoints", 0)

        coordinates.append(self.start.coords)

        if n_points > 2:

            for i in range(n_points - 2):
                coord = self._get_intermediate_coord((i + 1) / (n_points - 2 + 1))
                coordinates.append(coord)

        coordinates.append(self.end.coords)

        return coordinates

    def to_geojson(self):
        return self.linestring



from typing import Any, Sequence, Callable, Set, List, Union

from turf.helpers import (
    Feature,
    Point,
    LineString,
    MultiPoint,
    MultiLineString,
    Polygon,
    MultiPolygon,
    FeatureCollection,
    get_input_dimensions,
)
from turf.utils.exceptions import InvalidInput
from turf.utils.error_codes import error_code_messages
from turf.utils.helpers import dimensions

allowed_types_points = ["Point", "MultiPoint"]
allowed_types_default = [
    "Point",
    "MultiPoint",
    "LineString",
    "MultiLineString",
    "Polygon",
    "MultiPolygon",
]


def _process_list_input(
    allowed_types: Sequence,
    input_value: Sequence,
    callback: Callable,
    raise_exception: bool = True,
) -> List:
    """
    Retrieves the coords from lists, or raises an exception if invalid.

    :param allowed_types: allowed Feature types
    :param input_value: list input
    :param callback: method to call recursively
    :param raise_exception: if an exception should be raised or if it should be silent
    :return: list with extracted coords
    """

    dim = get_input_dimensions(input_value)

    allowed_dimensions = {dimensions.get(type_, None) for type_ in allowed_types}

    allowed_class_types = (
        *[eval(allowed_type) for allowed_type in allowed_types],
        dict,
    )

    if dim == 1 and all(isinstance(el, allowed_class_types) for el in input_value):
        return list(map(lambda geom: callback(geom, allowed_types), input_value))

    return _check_input_in_allowed_dimensions(
        input_value, dim, allowed_types, allowed_dimensions, raise_exception
    )


def _check_input_in_allowed_dimensions(
    input_value: Sequence,
    dim: int,
    allowed_types: Sequence,
    allowed_dimensions: Set,
    raise_exception: bool = True,
) -> Union[List, Sequence]:
    """
    :param input_value: list input
    :param dim: number of dimensions of input
    :param allowed_types: allowed Feature types
    :param allowed_dimensions: allowed dimensions for specified allowed types
    :param raise_exception: if an exception should be raised or if it should be silent
    :return: list with extracted coords
    """
    if dim in allowed_dimensions:
        if dim == 1 and len(input_value) < 2:
            raise InvalidInput(error_code_messages["InvalidPointInput"])

        return input_value
    else:
        if raise_exception:
            raise InvalidInput(
                error_code_messages["InvalidGeometry"](
                    [type_ for type_ in allowed_types if type_ in allowed_types_default]
                )
            )
        return []


def get_coords_from_geometry(
    geometry: Any, allowed_types: Sequence = None, raise_exception: bool = True
) -> List:
    """
    Retrieves coords from a given Geometry. Geometry must be a GeoJSON,
    a Geometry object or a list of coordinates, otherwise it raises an exception.

    :param geometry: Any input value(s)
    :param allowed_types: allowed Feature types
    :param raise_exception: if an exception should be raised or if it should be silent
    :return: list with extracted coords
    """

    if not allowed_types:
        allowed_types = allowed_types_default

    if isinstance(geometry, list):
        return _process_list_input(
            allowed_types, geometry, get_coords_from_geometry, raise_exception
        )

    if isinstance(geometry, (Feature, dict)):
        if geometry.get("type") == "Feature":
            return get_coords_from_geometry(geometry.get("geometry", {}), allowed_types)

    allowed_class_types = [
        *[eval(allowed_type) for allowed_type in allowed_types],
        dict,
    ]

    if any(
        isinstance(geometry, allowed_type_class)
        and geometry.get("type", "") in allowed_types
        for allowed_type_class in allowed_class_types
    ):
        return geometry.get("coordinates", [])

    if raise_exception:
        raise InvalidInput(error_code_messages["InvalidGeometry"](allowed_types))
    else:
        return []


def get_geometry_from_features(features: Any, allowed_types: Sequence = None) -> List:
    """
    Retrieves Geometries from Features. Features must be a GeoJSON,
    a Feature object or a list of coordinates, otherwise it raises an exception.

    :param features: Any input value(s)
    :param allowed_types: allowed Feature types
    :return: list with extracted coords
    """

    if not allowed_types:
        allowed_types = allowed_types_default

    if isinstance(features, list):
        return _process_list_input(
            [*allowed_types, "Feature"], features, get_geometry_from_features
        )

    if isinstance(features, (FeatureCollection, dict)):
        if features.get("type") == "FeatureCollection":
            return list(
                map(
                    lambda feature: feature.get("geometry", {}),
                    features.get("features", []),
                )
            )

    if isinstance(features, (Feature, dict)):
        if features.get("type", "") == "Feature":
            if features.get("geometry", {}).get("type", "") in allowed_types:
                return features.get("geometry", {})

    if isinstance(
        features, (*[eval(allowed_type) for allowed_type in allowed_types], dict)
    ):
        if features.get("type", "") in allowed_types:
            return features

    raise InvalidInput(error_code_messages["InvalidGeometry"](allowed_types))


def get_coords_from_features(features: Any, allowed_types: Sequence = None) -> List:
    """
    Retrieves coords from Features. Features must be a GeoJSON,
    a Feature object or a list of coordinates, otherwise it raises an exception.

    :param features: Any input value(s)
    :param allowed_types: allowed Feature types
    :return: list with extracted coords
    """

    if not allowed_types:
        allowed_types = allowed_types_default

    if isinstance(features, (FeatureCollection, dict)):
        if features.get("type") == "FeatureCollection":
            return list(
                map(
                    lambda feature: get_coords_from_geometry(
                        feature.get("geometry", {}), allowed_types
                    ),
                    features.get("features", []),
                )
            )

    return get_coords_from_geometry(features, allowed_types)



from turf.bbox import bbox
from turf.bbox_polygon import bbox_polygon


def envelope(features, *args):
    """
    Takes any number of features and returns a rectangular Polygon that encompasses all vertices.

    :param features: any GeoJSON feature or feature collection
    :return: bounding box extent in [minX, minY, maxX, maxY] order
    """

    return bbox_polygon(bbox(features, *args))



import json
import os
from collections import defaultdict


def get_fixtures(current_path, fixtures=None, keys=None):

    if not fixtures:
        fixtures = defaultdict(lambda: {"in": None, "out": None})

    if not keys:
        keys = ["in", "out"]

    for key in keys:

        files_path = os.path.join(current_path, key)

        for filename in os.listdir(files_path):
            with open(os.path.join(files_path, filename), "r") as f:
                name = ".".join(filename.split(".")[:-1])
                fixtures[name][key] = json.load(f)

    return fixtures



import numpy as np

from turf.invariant import get_coords_from_features
from turf.helpers import degrees_to_radians, radians_to_degrees


def bearing(start, end, options=None):
    """
    Takes two points and finds the geographic bearing between them,
    i.e. the angle measured in degrees from the north line (0 degrees)

    :param start: starting point [lng, lat] or Point feature
    :param end: ending point [lng, lat] or Point feature
    :param options: dictionary with options:
        [options["final"]] - calculates the final bearing if true
    :return: bearing in decimal degrees, between -180 and 180 (positive clockwise)
    """

    if not options:
        options = {}

    if isinstance(options, dict) and "final" in options:
        return calculate_final_bearing(start, end)

    start = get_coords_from_features(start, ["Point"])
    end = get_coords_from_features(end, ["Point"])

    lon1 = degrees_to_radians(start[0])
    lon2 = degrees_to_radians(end[0])
    lat1 = degrees_to_radians(start[1])
    lat2 = degrees_to_radians(end[1])

    a = np.sin(lon2 - lon1) * np.cos(lat2)

    b = np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * np.cos(lat2) * np.cos(lon2 - lon1)

    return radians_to_degrees(np.arctan2(a, b))


def calculate_final_bearing(start, end):
    bear = bearing(end, start)
    return (bear + 180) % 360



def get_input_dimensions(lst, n_dim=0):
    if isinstance(lst, list):
        return get_input_dimensions(lst[0], n_dim + 1) if len(lst) > 0 else 0
    else:
        return n_dim


def truncate(n, decimals=0):
    multiplier = 10 ** decimals
    return int(n * multiplier) / multiplier


dimensions = {
    "Point": 1,
    "MultiPoint": 2,
    "LineString": 2,
    "MultiLineString": 3,
    "Polygon": 3,
    "MultiPolygon": 4,
}



import numpy as np
from functools import reduce

from turf.invariant import get_geometry_from_features, get_coords_from_geometry
from turf.helpers import (
    earth_radius,
    degrees_to_radians as rad,
    get_input_dimensions,
    polygon,
)


def area(features):
    """
    Takes one or more features and returns their area in square meters.

    :param features: geojson input GeoJSON feature(s)
    :return: area in square meters
    """

    geometries = get_geometry_from_features(features)

    if not isinstance(geometries, list) or geometries == features:
        geometries = [geometries]

    return reduce(lambda prev, curr: prev + calculate_area(curr), geometries, 0)


def calculate_area(geometry):

    """
    Calculate geometry area

    :param geometry: GeoJSON geometry
    :return: the geometry area
    """

    coords = get_coords_from_geometry(
        geometry, ["Polygon", "MultiPolygon"], raise_exception=False
    )

    if get_input_dimensions(coords) >= 4:
        areas = list(map(lambda sub_item: calculate_area(sub_item), coords))
        return sum(areas)

    elif get_input_dimensions(coords) == 3:
        polygon(coords)
        return polygon_area(coords)

    else:
        return 0


def polygon_area(coords):
    """
    Calculates the area of a Polygon

    :param coords: the array of rings defining a Polygon
    :return: the total area of the Polygon
    """
    total = 0

    if len(coords) > 0:
        total += abs(ring_area(coords[0]))

    for i in coords[1:]:
        total -= abs(ring_area(coords[i]))

    return total


def ring_area(coords):
    """
    Calculate the approximate area of the polygon were it projected onto the earth.
    Note that this area will be positive if ring is oriented clockwise, otherwise it will be negative.

    Reference:
    Robert. G. Chamberlain and William H. Duquette, "Some Algorithms for Polygons on a Sphere",
    JPL Publication 07-03, Jet Propulsion
    Laboratory, Pasadena, CA, June 2007 https://trs.jpl.nasa.gov/handle/2014/40409

    :param coords: the array of coordinates defining a Polygon LinearRing
    :return: The approximate signed geodesic area of the polygon in square meters.
    """

    total = 0

    coords_length = len(coords)

    if coords_length > 2:

        for i in range(coords_length):
            if i == coords_length - 2:
                lower_index = coords_length - 2
                middle_index = coords_length - 1
                upper_index = 0

            elif i == coords_length - 1:
                lower_index = coords_length - 1
                middle_index = 0
                upper_index = 1

            else:
                lower_index = i
                middle_index = i + 1
                upper_index = i + 2

            p1 = coords[lower_index]
            p2 = coords[middle_index]
            p3 = coords[upper_index]

            total += (rad(p3[0]) - rad(p1[0])) * np.sin(rad(p2[1]))

        total = total * earth_radius ** 2 / 2

    return total



import numpy as np
from turf.helpers import degrees_to_radians, radians_to_length
from turf.invariant import get_coords_from_features


def distance(start, end, options=None):
    """
    Calculates the distance between two Points in degrees, radians, miles, or kilometers.
    This uses the [Haversine formula](http://en.wikipedia.org/wiki/Haversine_formula) to account for global curvature.

    :param start: starting point [lng, lat] or Point feature
    :param end: ending point [lng, lat] or Point feature
    :param options: dictionary with units as an attribute. Can be degrees, radians, miles, or kilometers
    :return: distance between the 2 points
    """

    kwargs = {}
    if isinstance(options, dict) and "units" in options:
        kwargs.update(options)

    coordinates1 = get_coords_from_features(start, ["Point"])
    coordinates2 = get_coords_from_features(end, ["Point"])

    d_lat = degrees_to_radians(coordinates2[1] - coordinates1[1])
    d_lon = degrees_to_radians(coordinates2[0] - coordinates1[0])

    lat1 = degrees_to_radians(coordinates1[1])
    lat2 = degrees_to_radians(coordinates2[1])

    distance_rad = calculate_radians_distance(d_lon, d_lat, lat1, lat2)

    return radians_to_length(distance_rad, **kwargs)


def calculate_radians_distance(dif_lon, dif_lat, lat1, lat2):
    """
    Calculates the distance between start and end

    basic haversine formula
    http://www.edwilliams.org/avform.htm#Dist
    https://en.wikipedia.org/wiki/Great-circle_distance

    :param dif_lon: longitudinal difference (radians) between start and ending points
    :param dif_lat: latitudinal difference (radians) between start and ending points
    :param lat1: radians latitude for starting point
    :param lat2: radians latitude for ending point

    :return: distance_radians
    """
    d = np.sin(dif_lat / 2) ** 2 + np.sin(dif_lon / 2) ** 2 * np.cos(lat1) * np.cos(
        lat2
    )
    d = 2 * np.arctan2(np.sqrt(d), np.sqrt(1 - d))

    return d



import numpy as np

from turf.helpers import (
    degrees_to_radians,
    length_to_radians,
    radians_to_degrees,
    point,
)
from turf.invariant import get_coords_from_features
from turf.utils.helpers import truncate


def destination(origin, distance, bearing, options=None):
    """
    Takes a Point and calculates the location of a destination point given a distance in
    degrees, radians, miles, or kilometers; and bearing in degrees.
    This uses the [Haversine formula](http://en.wikipedia.org/wiki/Haversine_formula) to account for global curvature.

    :param origin: starting point
    :param distance: distance from the origin point
    :param bearing: bearing ranging from -180 to 180
    :param options: optional parameters
        [options["units"]='kilometers'] miles, kilometers, degrees, or radians
        [options["properties"]={}] Translate properties to Point
    :return: destination GeoJSON Point feature
    """

    if not options:
        options = {}

    kwargs = {}
    if "units" in options:
        kwargs["units"] = options.get("units")

    coords = get_coords_from_features(origin, ["Point"])

    longitude1 = degrees_to_radians(coords[0])
    latitude1 = degrees_to_radians(coords[1])
    bearing_rads = degrees_to_radians(bearing)

    radians = length_to_radians(distance, **kwargs)

    latitude2 = np.arcsin(
        np.sin(latitude1) * np.cos(radians)
        + np.cos(latitude1) * np.sin(radians) * np.cos(bearing_rads)
    )

    longitude2 = longitude1 + np.arctan2(
        np.sin(bearing_rads) * np.sin(radians) * np.cos(latitude1),
        np.cos(radians) - np.sin(latitude1) * np.sin(latitude2),
    )

    lng = truncate(radians_to_degrees(longitude2), 6)
    lat = truncate(radians_to_degrees(latitude2), 6)

    return point([lng, lat], options.get("properties", None))



from turf.helpers import point
from turf.bbox import bbox


def center(features, options=None):
    """
    Takes a Feature or FeatureCollection and returns the absolute center point of all features.

    :param features: features or collection of features
    :param options: optional parameters
        [options["properties"]={}] Translate GeoJSON Properties to Point
        [options["bbox"]={}] Translate GeoJSON BBox to Point
        [options["id"]={}] Translate GeoJSON Id to Point
    :return: a Point feature at the absolute center point of all input features
    """

    if not options or not isinstance(options, dict):
        options = {}

    bounding_box = bbox(features)
    x = (bounding_box[0] + bounding_box[2]) / 2
    y = (bounding_box[1] + bounding_box[3]) / 2

    return point([x, y], options.get("properties", {}), options)



from functools import reduce

from turf.distance import distance
from turf.invariant import get_coords_from_features


def length(features, options=None):
    """
    Calculates the total length of the input Feature / FeatureCollection in the specified units.

    :param features: a Feature / FeatureCollection of types LineString, MultiLineString, Polygon or MultiPolygon
    :param options: optional parameters
        [options["units"]=kilometers] can be degrees, radians, miles, or kilometers
    :return: the measured distance
    """

    coords = get_coords_from_features(
        features, ["LineString", "MultiLineString", "Polygon", "MultiPolygon",]
    )

    if any(isinstance(inner_item, list) for item in coords for inner_item in item):
        distances = list(map(lambda sub_item: length(sub_item, options), coords))
        return sum(distances)

    total_distance = reduce(
        lambda accum, coord: accum + distance(coord[0], coord[1], options),
        zip(coords, coords[1:]),
        0,
    )

    return total_distance



error_codes = {
    "InvalidPoint": "InvalidInput",
    "InvalidLinePolygon": "InvalidInput",
    "InvalidPointInput": "InvalidInput",
}

error_code_corpus = {
    "InvalidUnits": lambda units: f"'{units}' is not a valid unit.",
    "InvalidLength": "<length> must be a positive number",
    "InvalidArea": "<area> must be a positive number",
    "InvalidDistance": "<distance> must be a positive number",
    "InvalidDegrees": "<degrees> must be a number",
    "InvalidRadians": "<radians> must be a number",
    "InvalidGeometry": lambda geometries: f"Input must be of type {', '.join(geometries)}",
    "InvalidFeaturesInput": "Input must be a FeatureCollection or Feature",
    "InvalidGeometryInput": "Input must be a valid Geometry",
    "InvalidFeature": "Input must be a valid Feature",
    "InvalidFeatureCollection": "Input must be a an array of Features",
    "InvalidGreatCircle": lambda start, end: f"GreatCircle can't be calculated between {start} and {end}",
    "InvalidGreatCirclePoints": lambda start, end: f"Input {start} and {end} are diametrically opposite, thus there is no single route but rather infinite",
    "InvalidPoint": "Input must be a Point geoJSON feature or an array of numbers.",
    "InvalidMultiInput": "Input coordinates must be an array ",
    "InvalidPointInput": "Input coordinates must be an array of 2 valid numbers.",
    "InvalidLineStringInput": "Input coordinates must be an array of valid Points.",
    "InvalidLinePoints": "Input coordinates must be an array of 2 or more valid Points.",
    "InvalidPolygonInput": "Input coordinates must be an array of valid rings.",
    "InvalidLinearRing": "Each Polygon ring coordinates must be a list of 4 or more Points.",
    "InvalidFirstLastPoints": "First and last Points of Polygon ring are not equivalent.",
    "InvalidBoundingBox": "The input bounding box must be an array of size 4",
    "InvalidCoordinates": "The input geometry(s) must have a coordinates attribute",
}

error_code_messages = {
    "InvalidUnits": error_code_corpus["InvalidUnits"],
    "InvalidLength": error_code_corpus["InvalidLength"],
    "InvalidArea": error_code_corpus["InvalidArea"],
    "InvalidDistance": error_code_corpus["InvalidDistance"],
    "InvalidDegrees": error_code_corpus["InvalidDegrees"],
    "InvalidRadians": error_code_corpus["InvalidRadians"],
    "InvalidGeometry": error_code_corpus["InvalidGeometry"],
    "InvalidFeaturesInput": error_code_corpus["InvalidFeaturesInput"],
    "InvalidGeometryInput": error_code_corpus["InvalidGeometryInput"],
    "InvalidGreatCircle": error_code_corpus["InvalidGreatCircle"],
    "InvalidGreatCirclePoints": error_code_corpus["InvalidGreatCirclePoints"],
    "InvalidFeature": error_code_corpus["InvalidFeature"],
    "InvalidFeatureCollection": error_code_corpus["InvalidFeatureCollection"],
    "InvalidPoint": error_code_corpus["InvalidPoint"],
    "InvalidPointInput": error_code_corpus["InvalidPointInput"],
    "InvalidMultiInput": error_code_corpus["InvalidMultiInput"],
    "InvalidLineStringInput": error_code_corpus["InvalidLineStringInput"],
    "InvalidLinePoints": error_code_corpus["InvalidLinePoints"],
    "InvalidPolygonInput": error_code_corpus["InvalidPolygonInput"],
    "InvalidLinearRing": error_code_corpus["InvalidLinearRing"],
    "InvalidFirstLastPoints": error_code_corpus["InvalidFirstLastPoints"],
    "InvalidBoundingBox": error_code_corpus["InvalidBoundingBox"],
    "InvalidCoordinates": error_code_corpus["InvalidCoordinates"],
}



from turf.invariant import get_coords_from_features
from turf.great_circle._arc import GreatCircle


def great_circle(start, end, options=None):
    """
    Returns  the great circle route as LineString

    :param start: source point feature
    :param end: destination point feature
    :param options: Optional parameters
        [options["properties"]={}] line feature properties
        [options.npoints=100] number of points
    :return: great circle line feature
    """
    if not options or not isinstance(options, dict):
        options = {}

    start = get_coords_from_features(start, ["Point"])
    end = get_coords_from_features(end, ["Point"])

    properties = options.get("properties", {})
    npoints = options.get("npoints", 100)
    properties["npoints"] = npoints

    gc = GreatCircle(start, end, properties)

    return gc.to_geojson()



import os
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    version = os.environ["GITHUB_REF"].split("/")[-1]
    print(f"Version: {version}")
except KeyError:
    try:
        with open("version.txt") as f:
            __version__ = f.read()
        version = __version__
    except FileNotFoundError:
        version = 0

setup(
    name="pyturf",
    version=version,
    description="Python geospatial library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/diogomatoschaves/py-turf",
    author="Diogo Matos Chaves",
    author_email="di.matoschaves@gmail.com",
    packages=[*find_packages(), "turf.utils"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5",
)



from turf.along import along
from turf.area import area
from turf.bbox import bbox
from turf.bbox_polygon import bbox_polygon
from turf.bearing import bearing
from turf.center import center
from turf.centroid import centroid
from turf.destination import destination
from turf.distance import distance
from turf.envelope import envelope
from turf.great_circle import great_circle
from turf.helpers import *
from turf.length import length
from turf.midpoint import midpoint
from turf.rhumb_bearing import rhumb_bearing
from turf.rhumb_destination import rhumb_destination
from turf.rhumb_distance import rhumb_distance
from turf.square import square
